// Copyright (c) 2020 CRS4
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#pragma once
#include <functional>
#include <pybind11/pybind11.h>
#include <eddl/utils.h>


void bind_eddl_utils(std::function<pybind11::module &(std::string const &namespace_)> &M) {

    pybind11::enum_<WrappingMode>(M(""), "WrappingMode", pybind11::arithmetic(), "")
	.value("Constant", WrappingMode::Constant)
	.value("Reflect", WrappingMode::Reflect)
	.value("Nearest", WrappingMode::Nearest)
	.value("Mirror", WrappingMode::Mirror)
	.value("Wrap", WrappingMode::Wrap)
	.value("Original", WrappingMode::Original)
	;

}
